/*-
 * Copyright (c) 2009 Andriy Gapon
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/uio.h>
#include <sys/fcntl.h>
#include <sys/ioccom.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/proc.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/condvar.h>
#include <sys/selinfo.h>
#include <sys/poll.h>

#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h>

#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>

#include <dev/heci/hecireg.h>
#include <dev/heci/heciio.h>
#include <dev/heci/heci.h>

/*
 * XXX This is maximum physical messages size, a logical message
 * can consist of multiple physical messages.
 * Calculated as maximum possible circular buffer size in bytes.
 */
#define	MAX_MSG_SIZE	(128 * 4)

#define	HECI_READ(sc, off)		bus_read_4(sc->res_memory, off)
#define	HECI_WRITE(sc, off, val)	bus_write_4(sc->res_memory, off, val)
#define	HECI_LOCK(sc)			mtx_lock(&(sc)->mtx)
#define	HECI_UNLOCK(sc)			mtx_unlock(&(sc)->mtx)
#define	CLIENT_LOCK(mec)		mtx_lock(&(mec)->mtx)
#define	CLIENT_UNLOCK(mec)		mtx_unlock(&(mec)->mtx)

#define	BZERO_OBJ(o)	memset(&o, 0, sizeof(o))
#define	BZERO_PTR(p)	memset(p, 0, sizeof(*p))
#define	HECI_CURRENTPID	(curthread->td_proc ? curthread->td_proc->p_pid : 0)

#define	RDPTR(val)		(((val) & H_CBRP) >> 8)
#define	WRPTR(val)		(((val) & H_CBWP) >> 16)
#define	DEPTH(val)		(((val) & H_CBD) >> 24)
#define	RD_SLOT_CNT(val)	(WRPTR(val) - RDPTR(val))
#define	WR_SLOT_CNT(val)	(DEPTH(val) - RD_SLOT_CNT(val))
#define	SLOTS_NEEDED(bsize)	(((bsize) + 3) / 4)

#define	CDEV2SOFTC(dev)		((struct heci_softc *)(dev)->si_drv1)
#define	DEVICE2SOFTC(dev)	((struct heci_softc *)device_get_softc(dev))

MALLOC_DEFINE(M_HECI, "heci memory", "heci memory");

struct me_client;
struct heci_softc;

struct host_client {
	struct heci_softc		*sc;
	struct me_client		*peer;
	int				flow_ctrl;
	uint8_t				addr;
};

struct me_client {
	SLIST_ENTRY(me_client)		next;
	struct heci_client_properties	props;
	struct host_client		*peer;
	uint8_t				addr;
	struct mtx			mtx;
	struct cv			cv;
	int				rcount;
	uint8_t				rbuf[MAX_MSG_SIZE];
	uint8_t				wbuf[MAX_MSG_SIZE];
	struct selinfo			si;
};

struct heci_softc {
	int				rid_memory;
	int				rid_irq;
	struct resource			*res_memory;
	struct resource			*res_irq;
	void				*irq_handle;
	int				msi_count;
	struct intr_config_hook		intr_hook;
	device_t			dev;
	struct cdev			*cdev;
	struct mtx			mtx;
	struct cv			cv;
	SLIST_HEAD(, me_client)		me_clients;
	struct unrhdr			*host_addr_unr;
	union hbm_any_request		bus_req;
	union hbm_any_response		bus_resp;
	int				waiting;
	int				served;
	int				faulted;
	int				reset;
};

static void		heci_startup(void *sc);
static void		heci_intr(void *arg);
static int		heci_bus_wait(struct heci_softc *sc);
static void		heci_reset(struct heci_softc *sc);
static int		heci_init_bus(struct heci_softc *sc);
static void		print_guid(const struct heci_guid *g);
static void		print_fw_client_props(struct heci_softc *sc,
    const struct heci_client_properties *props);

static d_open_t		heci_open;
static d_close_t	heci_close;
static d_ioctl_t	heci_ioctl;
static d_read_t		heci_read;
static d_write_t	heci_write;
static d_poll_t		heci_poll;

static struct
cdevsw	heci_cdevsw = {
	.d_version	= D_VERSION,
	.d_open		= heci_open,
	.d_close	= heci_close,	/* XXX can be dropped? */
	.d_ioctl	= heci_ioctl,
	.d_read		= heci_read,
	.d_write	= heci_write,
	.d_poll		= heci_poll,
	.d_name		= "heci"
};

static devclass_t	heci_devclass;

static int	heci_probe(device_t);
static int	heci_attach(device_t);
static int	heci_detach(device_t);

static device_method_t heci_methods[] = {
	DEVMETHOD(device_probe,		heci_probe),
	DEVMETHOD(device_attach,	heci_attach),
	DEVMETHOD(device_detach,	heci_detach),
	{ 0, 0 }
};

static driver_t	heci_driver = {
	"heci",
	heci_methods,
	sizeof(struct heci_softc)
};

DRIVER_MODULE(heci, pci, heci_driver, heci_devclass, 0, 0);
MODULE_DEPEND(heci, pci, 1, 1, 1);
MODULE_VERSION(heci, 1);


static int
heci_probe(device_t device)
{
	switch (pci_get_devid(device)) {
	case HECI_DEV_ID_ICH9_82G33:
		device_set_desc(device,
		    "Intel 82G33/G31/P35/P31 Express HECI/MEI Controller");
		return (0);
	case HECI_DEV_ID_ICH9M_1:
		device_set_desc(device,
		    "Intel ICH9 (Cantiga) HECI/MEI Controller");
		return (0);
	}
	return (ENXIO);
}

static int
heci_attach(device_t device)
{
	struct heci_softc	*sc = DEVICE2SOFTC(device);
	uint32_t		val;
	int			error;

	sc->dev = device;
	sc->rid_memory = HECI_MBAR;
	sc->rid_irq = 0;

	SLIST_INIT(&sc->me_clients);
	mtx_init(&sc->mtx, device_get_nameunit(device), "heci", MTX_DEF);
	cv_init(&sc->cv, "hecicv");

	/* Note: addresses 0, 1, 2 are reserved. */
	sc->host_addr_unr = new_unrhdr(3, 255, 0);

	sc->msi_count = pci_msi_count(device);
	if (sc->msi_count > 0) {
		sc->msi_count = 1;
		error = pci_alloc_msi(device, &sc->msi_count);
		if (error == 0) {
			device_printf(device, "using MSI\n");
			sc->rid_irq = 1;
		}
	}

	sc->res_irq = bus_alloc_resource_any(device, SYS_RES_IRQ, &sc->rid_irq,
	    RF_SHAREABLE | RF_ACTIVE);
	if (sc->res_irq == NULL) {
		device_printf(device, "failed to allocate irq resource\n");
		goto errexit;
	}

	sc->res_memory = bus_alloc_resource(device, SYS_RES_MEMORY,
	    &sc->rid_memory, 0ul, ~0ul, 16, RF_ACTIVE);
	if (sc->res_memory == NULL) {
		device_printf(device, "failed to allocate mem resource\n");
		goto errexit;
	}

	error = bus_setup_intr(device, sc->res_irq,
	    INTR_TYPE_MISC | INTR_MPSAFE, NULL, heci_intr, sc, &sc->irq_handle);
	if (error) {
		device_printf(device, "failed to set up interrupt handler\n");
		goto errexit;
	}

	/* Acknowledge old interrupt if any. */
	val = HECI_READ(sc, H_CSR);
	if (val & H_IS)
		HECI_WRITE(sc, H_CSR, val);

	heci_reset(sc);

	val = HECI_READ(sc, ME_CSR_HA);
	if ((val & ME_RDY_HRA) == 0) {
		device_printf(device, "ME is not ready after reset\n");
		goto errexit;
	}
	val = HECI_READ(sc, H_CSR);
	if ((val & H_RDY) == 0) {
		device_printf(device, "host is not ready after reset\n");
		goto errexit;
	}

	/* Defer remaining of initialization until interrupts are enabled */
	sc->intr_hook.ich_func = heci_startup;
	sc->intr_hook.ich_arg = sc;
	if (!cold) {
		sc->intr_hook.ich_func = NULL;
		heci_startup(sc);
	}
	else if (config_intrhook_establish(&sc->intr_hook) != 0) {
		device_printf(device, "config_intrhook_establish failed\n");
		goto errexit;
	}
	return (0);

errexit:
	heci_detach(device);
	return (ENXIO);
}

static int
heci_detach(device_t device)
{
	struct heci_softc	*sc = DEVICE2SOFTC(device);

	/* First, severe ties with userland. */
	if (sc->cdev)
		destroy_dev(sc->cdev);

	if (sc->irq_handle != NULL) {
		bus_teardown_intr(device, sc->res_irq, sc->irq_handle);
		sc->irq_handle = NULL;
	}
	if (sc->res_irq != NULL) {
		bus_deactivate_resource(device, SYS_RES_IRQ, sc->rid_irq,
		    sc->res_irq);
		bus_release_resource(device, SYS_RES_IRQ, sc->rid_irq,
		    sc->res_irq);
		sc->res_irq = NULL;
		if (sc->msi_count > 0)
			pci_release_msi(device);
	}
	if (sc->res_memory != NULL) {
		bus_deactivate_resource(device, SYS_RES_MEMORY, sc->rid_memory,
		    sc->res_memory);
		bus_release_resource(device, SYS_RES_MEMORY, sc->rid_memory,
		    sc->res_memory);
		sc->res_memory = NULL;
	}

	while (!SLIST_EMPTY(&sc->me_clients)) {
		struct me_client *mec = SLIST_FIRST(&sc->me_clients);

		SLIST_REMOVE_HEAD(&sc->me_clients, next);
		cv_destroy(&mec->cv);
		mtx_destroy(&mec->mtx);
		free(mec, M_HECI);
	}
	delete_unrhdr(sc->host_addr_unr);
	cv_destroy(&sc->cv);
	mtx_destroy(&sc->mtx);

	return (0);
}

static void
heci_startup(void *arg)
{
	struct heci_softc	*sc = arg;
	int			unit = device_get_unit(sc->dev);

	/* Remove ourselves from the config hooks */
	if (sc->intr_hook.ich_func != NULL) {
		config_intrhook_disestablish(&sc->intr_hook);
		sc->intr_hook.ich_func = NULL;
	}

	if (heci_init_bus(sc) != 0) {
		sc->faulted = 1;	/* XXX detach?! */
		return;
	}

	sc->cdev = make_dev(&heci_cdevsw, 0, UID_ROOT, GID_OPERATOR, 0600,
	    "heci%d", unit);
	sc->cdev->si_drv1 = sc;
}

static void
heci_reset(struct heci_softc *sc)
{
	uint32_t val;
	int	err = 0;

	HECI_LOCK(sc);

	val = HECI_READ(sc, H_CSR);
	val |= (H_RST | H_IG);
	val |= H_IE;
	HECI_WRITE(sc, H_CSR, val);

	DELAY(100);
	val = HECI_READ(sc, H_CSR);
	val |= (H_IG | H_RDY);
	val &= ~H_RST;
	HECI_WRITE(sc, H_CSR, val);

	/* wait for initial interrupt from ME */
	sc->reset = 1;
	if (!cold)
		err = cv_timedwait(&sc->cv, &sc->mtx, 7 * hz);	/*XXX*/
	else
		DELAY(100);
	sc->reset = 0;
	if (err)
		device_printf(sc->dev, "timeout waiting for reset interrupt\n");

	if (bootverbose)
		device_printf(sc->dev,
		    "after rest H_CSR=0x%08x, ME_CSR_HA=0x%08x\n",
		    HECI_READ(sc, H_CSR),
		    HECI_READ(sc, ME_CSR_HA));

	HECI_UNLOCK(sc);
}

static void
heci_fault(struct heci_softc *sc)
{
	struct me_client *mec;

	sc->faulted = 1;

	/* Unblock userland. */
	SLIST_FOREACH(mec, &sc->me_clients, next) {
		selwakeup(&mec->si);
		cv_signal(&mec->cv);
	}
}

static int
heci_write_cbuf(struct heci_softc *sc, struct heci_msg_hdr hdr,
    const uint8_t *data)
{
	const uint32_t	*wr_buf;
	uint32_t	tmp;
	uint32_t	val;
	uint32_t	to_write;

	mtx_assert(&sc->mtx, MA_OWNED);

	val = HECI_READ(sc, ME_CSR_HA);
	if ((val & ME_RDY_HRA) == 0) {
		device_printf(sc->dev, "heci write: ME not ready\n");
		heci_fault(sc);
		return (ENXIO);
	}

	wr_buf = (const uint32_t *)data;
	to_write = hdr.length;

	val = HECI_READ(sc, H_CSR);
	if (WR_SLOT_CNT(val) < SLOTS_NEEDED(to_write)) {
		device_printf(sc->dev, "want to write %d, available %d\n",
		    SLOTS_NEEDED(to_write), WR_SLOT_CNT(val));
		return (ENOBUFS);
	}

	memcpy(&tmp, &hdr, 4);
	HECI_WRITE(sc, H_CB_WW, tmp);

	while (to_write >= 4) {
		HECI_WRITE(sc, H_CB_WW, *wr_buf);
		to_write -= 4;
		wr_buf++;
	}
	if (to_write > 0) {
		tmp = 0;
		memcpy(&tmp, wr_buf, to_write);
		HECI_WRITE(sc, H_CB_WW, tmp);
	}

	val = HECI_READ(sc, H_CSR);
	HECI_WRITE(sc, H_CSR, (val & ~H_IS) | H_IG);
	return (0);
}

static void
heci_read_cbuf(struct heci_softc *sc, unsigned int size, uint8_t *buf)
{
	uint32_t	*rd_buf;
	uint32_t	val;

	mtx_assert(&sc->mtx, MA_OWNED);

	rd_buf = (uint32_t *)buf;

	while (size >= 4) {
		*rd_buf = HECI_READ(sc, ME_CB_RW);
		size -= 4;
		rd_buf++;
	}
	if (size > 0) {
		val = HECI_READ(sc, ME_CB_RW);
		memcpy(rd_buf, &val, size);
	}

	val = HECI_READ(sc, H_CSR);
	HECI_WRITE(sc, H_CSR, (val & ~H_IS) | H_IG);
}

static void
heci_drain_cbuf(struct heci_softc *sc, unsigned int count)
{
	uint32_t	val;

	mtx_assert(&sc->mtx, MA_OWNED);

	while (count >= 4) {
		HECI_READ(sc, ME_CB_RW);
		count -= 4;
	}
	if (count > 0)
		HECI_READ(sc, ME_CB_RW);

	val = HECI_READ(sc, H_CSR);
	HECI_WRITE(sc, H_CSR, (val & ~H_IS) | H_IG);
}

static int
heci_bus_wait(struct heci_softc *sc)
{
	int error = 0;

	mtx_assert(&sc->mtx, MA_OWNED);

	sc->served = 0;
	sc->waiting = 1;

	while (!sc->served && error == 0)
		error = cv_timedwait(&sc->cv, &sc->mtx, 7 * hz); /*XXX*/

	sc->waiting = 0;
	cv_signal(&sc->cv);

	switch (error)
	{
	case 0:
		if (sc->served == 0 && bootverbose)
			device_printf(sc->dev, "interrupt with no data\n");
		break;
	case EWOULDBLOCK:
		device_printf(sc->dev, "device timeout\n");
		break;
	default:
		device_printf(sc->dev, "error from cv_timedwait: %d\n", error);
		break;
	}

	return (error);
}

static void
heci_bus_release(struct heci_softc *sc)
{
	mtx_assert(&sc->mtx, MA_OWNED);

	if (!sc->served)
		device_printf(sc->dev,
		    "heci_bus_release called with served = 0\n");

	sc->served = 0;
	sc->waiting = 0;

	cv_signal(&sc->cv);
}

static void
handle_bus_msg(struct heci_softc *sc, struct heci_msg_hdr hdr)
{
	struct me_client	*mec;
	int			err = 0;

	/* A bus message? */
	if (hdr.host_addr != 0) {
		device_printf(sc->dev,
		    "impossible peers: host_addr=0x%02x, me_addr=0x%02x\n",
		    hdr.host_addr,
		    hdr.me_addr);
		heci_fault(sc);
		return;
	}
	if (hdr.length > sizeof(sc->bus_resp)) {
		device_printf(sc->dev,
		    "ME bus message is too long: %u\n",
		    hdr.length);
		heci_fault(sc);
		return;
	}
	if (hdr.msg_complete == 0) {
		device_printf(sc->dev, "ME bus message is not complete\n");
		heci_fault(sc);
		return;
	}

	/* Check if previous bus message is processed */
	while (sc->served && err == 0)
		err = cv_timedwait(&sc->cv, &sc->mtx, 7 * hz); /*XXX*/
	if (err) {
		device_printf(sc->dev,
		    "error waiting for previous bus response to get processed\n");
		heci_fault(sc);
		return;
	}

	heci_read_cbuf(sc, hdr.length, (uint8_t *)&sc->bus_resp);

	if (bootverbose)
		device_printf(sc->dev,
		    "bus message from ME: cmd	0x%x, is_response = %u\n",
		    sc->bus_resp.cmd.cmd,
		    sc->bus_resp.cmd.is_response);

	if (sc->bus_resp.cmd.is_response) {
		if (sc->waiting == 0) {
			device_printf(sc->dev,
			    "bus response but nobody's waiting\n");
			device_printf(sc->dev,
			    "cmd = 0x%x\n",
			    sc->bus_resp.cmd.cmd);
			return;
		}
		/* TODO Check if response size corresponds to response type. */
		sc->served = 1;
		cv_signal(&sc->cv);
		return;
	}

	switch (sc->bus_resp.cmd.cmd) {
	case ME_STOP_REQ_CMD:
		/* XXX How to handle? OpenAMT driver issue HOST_STOP_REQ_CMD */
		device_printf(sc->dev, "ME issued ME_STOP_REQ_CMD, ignored\n");
		break;
	case HECI_FLOW_CONTROL_CMD: {
		SLIST_FOREACH(mec, &sc->me_clients, next)
		if (mec->addr == sc->bus_resp.flow.me_addr)
			break;

		if (mec == NULL)
			device_printf(sc->dev,
			    "flow control from unknown ME addr 0x%02x\n",
			    sc->bus_resp.flow.me_addr);
		else if (mec->peer == NULL)
			device_printf(sc->dev,
			    "flow control from unconnected ME addr 0x%02x\n",
			    sc->bus_resp.flow.me_addr);
		else if (mec->peer->addr != sc->bus_resp.flow.host_addr)
			device_printf(sc->dev,
			    "flow control to mismatching host addr: ME addr 0x%02x, "
			    "connected to host addr 0x%02x, received 0x%02x\n",
			    sc->bus_resp.flow.me_addr,
			    mec->peer->addr,
			    sc->bus_resp.flow.host_addr);
		else {
			CLIENT_LOCK(mec);
			mec->peer->flow_ctrl++;
			if (mec->peer->flow_ctrl == 1) {
				selwakeup(&mec->si);
				cv_signal(&mec->cv);
			}
			CLIENT_UNLOCK(mec);
		}
		break;
	}
	case CLIENT_DISCONNECT_REQ_CMD: {
		if (bootverbose)
			device_printf(sc->dev,
			    "disconnect request from ME client\n");

		SLIST_FOREACH(mec, &sc->me_clients, next)
		if (mec->addr == sc->bus_resp.disconn.me_addr)
			break;

		if (mec == NULL)
			device_printf(sc->dev,
			    "disconnect from unknown ME addr 0x%02x\n",
			    sc->bus_resp.disconn.me_addr);
		else if (mec->peer == NULL)
			device_printf(sc->dev,
			    "disconnect from unconnected ME addr 0x%02x\n",
			    sc->bus_resp.disconn.me_addr);
		else if (mec->peer->addr != sc->bus_resp.flow.host_addr)
			device_printf(sc->dev,
			    "disconnect to mismatching host addr: ME addr 0x%02x, "
			    "connected to host addr 0x%02x, received 0x%02x\n",
			    sc->bus_resp.disconn.me_addr,
			    mec->peer->addr,
			    sc->bus_resp.disconn.host_addr);
		else {
			struct heci_msg_hdr resp_hdr;
			struct hbm_client_disconnect_response resp;

			resp_hdr.host_addr = 0;
			resp_hdr.me_addr = 0;
			resp_hdr.msg_complete = 1;
			resp_hdr.reserved = 0;
			resp_hdr.length =
			    sizeof(struct hbm_client_disconnect_response);
			resp.cmd.cmd = CLIENT_DISCONNECT_REQ_CMD;
			resp.cmd.is_response = 1;
			resp.me_addr = mec->addr;
			resp.host_addr = mec->peer->addr;
			resp.status = 0;
			heci_write_cbuf(sc, resp_hdr, (const uint8_t *)&resp);

			CLIENT_LOCK(mec);
			mec->peer->peer = NULL;
			mec->peer = NULL;
			selwakeup(&mec->si);
			cv_signal(&mec->cv);
			CLIENT_UNLOCK(mec);
		}
		break;
	}
	default:
		device_printf(sc->dev,
		    "ME issued unknown command 0x%02x\n",
		    sc->bus_resp.cmd.cmd);
	}
}

static void
handle_app_msg(struct heci_softc *sc, struct heci_msg_hdr hdr)
{
	struct me_client *mec;

	/* An application message? */
	if (hdr.host_addr == 0) {
		device_printf(sc->dev,
		    "impossible peers: host_addr=0x%02x, me_addr=0x%02x\n",
		    hdr.host_addr, hdr.me_addr);
		heci_fault(sc);
		return;
	}

	/* XXX What to do with msg_complete?
	 * Ignoring for time being, just process
	 * all chunks as independent messages.
	 */
	SLIST_FOREACH(mec, &sc->me_clients, next)
	if (mec->addr == hdr.me_addr)
		break;

	if (mec == NULL) {
		device_printf(sc->dev,
		    "message from unknown ME addr 0x%02x\n",
		    hdr.me_addr);
		heci_drain_cbuf(sc, hdr.length);
	}
	else if (mec->peer == NULL) {
		device_printf(sc->dev,
		    "flow control from unconnected ME addr 0x%02x\n",
		    hdr.me_addr);
		heci_drain_cbuf(sc, hdr.length);
	}
	else if (mec->peer->addr != hdr.host_addr) {
		device_printf(sc->dev,
		    "message to mismatching host addr: ME addr 0x%02x, "
		    "connected to host addr 0x%02x, received 0x%02x\n",
		    hdr.me_addr, mec->peer->addr, hdr.host_addr);
		heci_drain_cbuf(sc, hdr.length);
	}
	else {
		CLIENT_LOCK(mec);
		/* If there is unread msg in buffer, discard this one. */
		if (mec->rcount)
			heci_drain_cbuf(sc, hdr.length);
		else {
			heci_read_cbuf(sc, hdr.length, mec->rbuf);
			mec->rcount = hdr.length;
			cv_signal(&mec->cv);
			selwakeup(&mec->si);
		}
		CLIENT_UNLOCK(mec);
	}
}

static void
heci_intr(void *arg)
{
	struct heci_msg_hdr	hdr;
	struct heci_softc	*sc = (struct heci_softc *)arg;
	uint32_t		val;
	int			first = 1;

	HECI_LOCK(sc);

	val = HECI_READ(sc, H_CSR);
	if ((val & H_IS) == 0) {
		HECI_UNLOCK(sc);
		return;
	}
	HECI_WRITE(sc, H_CSR, val);

	for (;;) {
		val = HECI_READ(sc, ME_CSR_HA);
		if ((val & ME_RDY_HRA) == 0) {
			device_printf(sc->dev, "heci_intr: ME not ready\n");
			heci_fault(sc);
			break;
		}

		/* Check if anything is available for reading. */
		if (RD_SLOT_CNT(val) == 0) {
			/* Signal only if reset is in progress, otherwise ME is
			 * telling us that it read data from circular buffer and
			 * is ready for more.
			 * XXX now we don't have any use for this information.
			 */
			if (first && sc->reset)
				cv_signal(&sc->cv);
			break;
		}
		first = 0;

		/* Read message header */
		val = HECI_READ(sc, ME_CB_RW);
		memcpy(&hdr, &val, 4);
		val = HECI_READ(sc, ME_CSR_HA);
		if (RD_SLOT_CNT(val) < SLOTS_NEEDED(hdr.length)) {
			/* XXX Is this too strict and fragile? */
			device_printf(sc->dev,
			    "msg header implies %d slots, available %d\n",
			    SLOTS_NEEDED(hdr.length),
			    RD_SLOT_CNT(val));
			heci_fault(sc);
			break;
		}

		if (hdr.me_addr == 0)
			handle_bus_msg(sc, hdr);
		else
			handle_app_msg(sc, hdr);
	}
	HECI_UNLOCK(sc);
}

static int
heci_bus_cmd(struct heci_softc *sc, int wait)
{
	struct heci_msg_hdr	req_hdr;
	int			resp_length;
	int			err = 0;

	mtx_assert(&sc->mtx, MA_OWNED);

	/* Wait until previous waiter gets a reply. */
	while (sc->waiting && err == 0)
		err = cv_timedwait(&sc->cv, &sc->mtx, 7 * hz); /*XXX*/
	if (err) {
		device_printf(sc->dev,
		    "error waiting for previous waiter to get served\n");
		heci_fault(sc);
		return (err);
	}

	req_hdr.host_addr = 0;
	req_hdr.me_addr = 0;
	req_hdr.msg_complete = 1;
	req_hdr.reserved = 0;

	if (bootverbose)
		device_printf(sc->dev,
		    "sending bus request command 0x%x\n",
		    sc->bus_req.cmd.cmd);

	switch (sc->bus_req.cmd.cmd) {
	case HOST_START_REQ_CMD:
		BZERO_OBJ(sc->bus_req.vers.reserved);
		req_hdr.length = sizeof(struct hbm_host_version_request);
		resp_length = sizeof(struct hbm_host_version_response);
		break;
	case HOST_STOP_REQ_CMD:
		BZERO_OBJ(sc->bus_req.stop.reserved);
		req_hdr.length = sizeof(struct hbm_host_stop_request);
		resp_length = sizeof(struct hbm_host_stop_response);
		break;
	case HOST_ENUM_REQ_CMD:
		BZERO_OBJ(sc->bus_req.enumer.reserved);
		req_hdr.length = sizeof(struct hbm_host_enum_request);
		resp_length = sizeof(struct hbm_host_enum_response);
		break;
	case HOST_CLIENT_PROPERTIES_REQ_CMD:
		BZERO_OBJ(sc->bus_req.props.reserved);
		req_hdr.length = sizeof(struct hbm_props_request);
		resp_length = sizeof(struct hbm_props_response);
		break;
	case CLIENT_CONNECT_REQ_CMD:
		BZERO_OBJ(sc->bus_req.conn.reserved);
		req_hdr.length = sizeof(struct hbm_client_connect_request);
		resp_length = sizeof(struct hbm_client_connect_response);
		break;
	case CLIENT_DISCONNECT_REQ_CMD:
		BZERO_OBJ(sc->bus_req.disconn.reserved);
		req_hdr.length = sizeof(struct hbm_client_disconnect_request);
		resp_length = sizeof(struct hbm_client_disconnect_response);
		break;
	case HECI_FLOW_CONTROL_CMD:
		if (wait) {
			device_printf(sc->dev,
			    "HECI_FLOW_CONTROL_CMD may not have response\n");
			return (EINVAL);
		}
		BZERO_OBJ(sc->bus_req.flow.reserved);
		req_hdr.length = sizeof(struct hbm_flow_control);
		break;
	default:
		device_printf(sc->dev,
		    "unknown bus request command 0x%x\n",
		    sc->bus_req.cmd.cmd);
		return (EINVAL);
	}

	err = heci_write_cbuf(sc, req_hdr, (const uint8_t *)&sc->bus_req);
	if (err) {
		device_printf(sc->dev,
		    "write error for bus request command 0x%x\n",
		    sc->bus_req.cmd.cmd);
		return (err);
	}

	if (!wait)
		return (0);

	err = heci_bus_wait(sc);
	if (err) {
		device_printf(sc->dev,
		    "wait error for bus request command 0x%x\n",
		    sc->bus_req.cmd.cmd);
		return (err);
	}
#if 0
	if (resp_hdr.length != resp_length) {
		device_printf(sc->dev,
		    "incorrect response size %u (expected %u)\n",
		    resp_hdr.length, resp_length);
		return EINVAL;
	}
#endif
	if (sc->bus_resp.cmd.cmd != sc->bus_req.cmd.cmd) {
		device_printf(sc->dev,
		    "mismatch between request and response types: %u vs %u\n",
		    sc->bus_req.cmd.cmd,
		    sc->bus_resp.cmd.cmd);
		return (EINVAL);
	}
	if (!sc->bus_resp.cmd.is_response) {
		device_printf(sc->dev, "response bit is not set in bus reply\n");
		return (EINVAL);
	}

	return (0);
}

static int
heci_init_bus(struct heci_softc *sc)
{
	uint8_t			valid_addresses[32];
	struct me_client	*mec;
	int			byte;
	int			bit;
	int			err;
	int			i;

	HECI_LOCK(sc);

	/* Version handshake. */
	sc->bus_req.vers.cmd.cmd = HOST_START_REQ_CMD;
	sc->bus_req.vers.host_version.major_version = HBM_MAJOR_VERSION;
	sc->bus_req.vers.host_version.minor_version = HBM_MINOR_VERSION;

	err = heci_bus_cmd(sc, 1);
	if (err) {
		device_printf(sc->dev,
		    "failed to establish bus communication\n");
		goto exit;
	}
	if (!sc->bus_resp.vers.host_version_supported) {
		device_printf(sc->dev,
		    "this driver API version is not supported\n");
		device_printf(sc->dev, "max supported version is %u.%u\n",
		    sc->bus_resp.vers.me_max_version.major_version,
		    sc->bus_resp.vers.me_max_version.minor_version);
		/* XXX issue ME STOP command */
		err = ENXIO;
		goto exit;
	}

	/* Enumerate FW clients. */
	sc->bus_req.enumer.cmd.cmd = HOST_ENUM_REQ_CMD;
	err = heci_bus_cmd(sc, 1);
	if (err) {
		device_printf(sc->dev, "failed to enumerate ME clients\n");
		goto exit;
	}
	memcpy(&valid_addresses[0],
		&sc->bus_resp.enumer.valid_addresses[0],
		sizeof(valid_addresses));

	for (i = 0; i < 256; i++) {
		byte = i / 8;
		bit = i % 8;

		if ((valid_addresses[byte] & (1 << bit)) == 0)
			continue;

		sc->bus_req.props.cmd.cmd = HOST_CLIENT_PROPERTIES_REQ_CMD;
		sc->bus_req.props.address = i;
		err = heci_bus_cmd(sc, 1);
		if (err) {
			device_printf(sc->dev,
			    "failed to query properties of fw client" \
			    "with addr 0x%02x\n", i);
			break;
		}

		device_printf(sc->dev, "found ME client at address 0x%02x:\n", i);
		
		/* XXX How do we interpret client status, just report for now. */
		device_printf(sc->dev, "\tstatus = 0x%02x\n",
		    sc->bus_resp.props.status);
		print_fw_client_props(sc,
		    &sc->bus_resp.props.client_properties);

		/* XXX What does "fixed address" mean?
		 * OpenAMT driver skips these.
		 */
		if (sc->bus_resp.props.client_properties.fixed_address)
			continue;
		if (sc->bus_resp.props.client_properties.max_number_of_connections == 0)
			continue;

		mec = malloc(sizeof(*mec), M_HECI, M_NOWAIT | M_ZERO);
		if (mec == NULL) {
			device_printf(sc->dev, "failed to allocate memory\n");
			err = ENOMEM;
			break;
		}
		mec->addr = i;
		memcpy(&mec->props, &sc->bus_resp.props.client_properties,
		    sizeof(mec->props));
		mtx_init(&mec->mtx, "heci_me", "heci", MTX_DEF);
		cv_init(&mec->cv, "hecimecv");
		SLIST_INSERT_HEAD(&sc->me_clients, mec, next);
	}
exit:
	heci_bus_release(sc);
	HECI_UNLOCK(sc);
	return (err);
}

static void
heci_client_dtr(void *data)
{
	struct host_client *hc = (struct host_client *)data;
	struct heci_softc	*sc = hc->sc;

	if (sc == NULL) {
		free(data, M_HECI);
		return;
	}

	/* Check if host client is connected. */
	if (hc->peer) {
		HECI_LOCK(sc);
		sc->bus_req.disconn.cmd.cmd = CLIENT_DISCONNECT_REQ_CMD;
		sc->bus_req.disconn.me_addr = hc->peer->addr;
		sc->bus_req.disconn.host_addr = hc->addr;

		heci_bus_cmd(sc, 1);
		heci_bus_release(sc);

		mtx_lock(&hc->peer->mtx);
		hc->peer->peer = NULL;
		hc->peer->rcount = 0;
		mtx_unlock(&hc->peer->mtx);
		HECI_UNLOCK(sc);
	}
	if (hc->addr)
		free_unr(sc->host_addr_unr, hc->addr);
	free(data, M_HECI);
}

static int
heci_open(struct cdev *cdev, int oflags, int devtype, struct thread *td)
{
	struct heci_softc	*sc = CDEV2SOFTC(cdev);
	struct host_client	*hc;
	int			err;
	uint8_t			haddr;

	if (sc == NULL) {
		/* Impossible. */
		printf("heci: null sc in open\n");
		return (EINVAL);
	}

	/* Must be opened RW, no other flags are supported. */
	if ((oflags & (FREAD | FWRITE)) != (FREAD | FWRITE)) {
		device_printf(sc->dev, "bad flags in open: 0x%x\n", oflags);
		return (EINVAL);
	}

	if (sc->faulted)
		return (ENXIO);

	haddr = alloc_unr(sc->host_addr_unr);
	if (haddr == 0) {
		/* all client addresses are in use */
		device_printf(sc->dev, "can not allocate host address\n");
		return (EBUSY);
	}

	hc = malloc(sizeof(*hc), M_HECI, M_WAITOK | M_ZERO);
	err = devfs_set_cdevpriv(hc, heci_client_dtr);
	if (err) {
		device_printf(sc->dev, "devfs_set_cdevpriv failed: %d\n", err);
		heci_client_dtr(hc);
		return (err);
	}
	hc->addr = haddr;
	hc->sc = sc;

	return (0);
}

static int
heci_close(struct cdev *cdev, int fflag, int devtype, struct thread *td)
{
	return (0);
}

static int
heci_read(struct cdev *cdev, struct uio *uio, int ioflag)
{
	struct heci_softc	*sc = CDEV2SOFTC(cdev);
	struct host_client	*hc;
	struct me_client	*mec;
	int			toread;
	int			err;

	if (ioflag) {
		/* XXX flags like FNONBLOCK are not supported yet. */
		device_printf(sc->dev, "bad flags in read: 0x%x\n", ioflag);
		return (EINVAL);
	}

	if (sc->faulted)
		return (ENXIO);

	err = devfs_get_cdevpriv((void **)&hc);
	if (err) {
		device_printf(sc->dev,
		    "heci_read: devfs_get_cdevpriv failed: %d\n", err);
		return (err);
	}
	mec = hc->peer;
	if (mec == NULL) {
		device_printf(sc->dev, "heci_read for unconnected client\n");
		return (ENOTCONN);
	}

	/* First, send flow control to ME peer */
	if (!mec->rcount) {
		HECI_LOCK(sc);
		sc->bus_req.flow.cmd.cmd = HECI_FLOW_CONTROL_CMD;
		sc->bus_req.flow.host_addr = hc->addr;
		sc->bus_req.flow.me_addr = mec->addr;
		err = heci_bus_cmd(sc, 0);
		HECI_UNLOCK(sc);
		if (err)
			return (err);
	}

	/* Wait until we get something into buffer */
	CLIENT_LOCK(mec);
	while (!mec->rcount && !sc->faulted && mec->peer == hc) {
		err = cv_timedwait_sig(&mec->cv, &mec->mtx, 7 * hz);
		if (err != 0) {
			if (err != EINTR)
				device_printf( sc->dev,
				    "heci_read: cv_timedwait_sig returned %d\n",
				    err);
			break;
		}
	}
	CLIENT_UNLOCK(mec);
	if (err)
		return (err);
	if (sc->faulted)
		return (ENXIO);
	if (mec->peer != hc)
		return (ECONNRESET);

	/* XXX What if userland tries to read less data than available in the buffer?
	 * For time being the remaining data is lost;
	 */
	toread = (min(uio->uio_resid, mec->rcount));
	err = uiomove(mec->rbuf, toread, uio);

	CLIENT_LOCK(mec);
	mec->rcount = 0;
	CLIENT_UNLOCK(mec);

	return (err);
}

static int
heci_write(struct cdev *cdev, struct uio *uio, int ioflag)
{
	struct heci_msg_hdr	hdr;
	struct heci_softc	*sc = CDEV2SOFTC(cdev);
	struct host_client	*hc;
	struct me_client	*mec;
	int			towrite;
	int			err;

	if (sc->faulted)
		return (ENXIO);

	if (ioflag) {
		/* XXX flags like FNONBLOCK are not supported yet. */
		device_printf(sc->dev, "bad flags in write: 0x%x\n", ioflag);
		return (EINVAL);
	}
	/* XXX What if userland wants to send a large (multipart message)?
	 * Refuse this for now;
	 * Also, should not forget about max_msg_length property.
	 */
	if (uio->uio_resid > sizeof(mec->wbuf))
		return (EINVAL);

	err = devfs_get_cdevpriv((void **)&hc);
	if (err) {
		device_printf(sc->dev,
		    "heci_write: devfs_get_cdevpriv failed: %d\n", err);
		return (err);
	}
	mec = hc->peer;
	if (mec == NULL) {
		device_printf(sc->dev, "heci_write for unconnected client\n");
		return (ENOTCONN);
	}

	/* Check if and wait until we have a permission to send to ME peer. */
	CLIENT_LOCK(mec);
	while (!hc->flow_ctrl && !sc->faulted && mec->peer == hc) {
		err = cv_timedwait_sig(&mec->cv, &mec->mtx, 7 * hz);
		if (err != 0) {
			if (err != EINTR)
				device_printf( sc->dev,
				    "heci_write: cv_timedwait_sig returned %d\n",
				    err);
			break;
		}
	}
	CLIENT_UNLOCK(mec);
	if (err)
		return (err);
	if (sc->faulted)
		return (ENXIO);
	if (mec->peer != hc)
		return (ECONNRESET);

	/*towrite = (min(uio->uio_resid, sizeof(mec->wbuf)));*/
	towrite = uio->uio_resid;
	err = uiomove(mec->wbuf, towrite, uio);
	if (err)
		return (err);

	hdr.host_addr = hc->addr;
	hdr.me_addr = mec->addr;
	hdr.msg_complete = 1;	/* XXX see comment above */
	hdr.reserved = 0;
	hdr.length = towrite;

	HECI_LOCK(sc);
	hc->flow_ctrl--;
	err = heci_write_cbuf(sc, hdr, mec->wbuf);
	HECI_UNLOCK(sc);

	return (err);
}

static int
heci_poll(struct cdev *cdev, int events, struct thread *td)
{
	struct heci_softc	*sc = CDEV2SOFTC(cdev);
	struct host_client	*hc;
	struct me_client	*mec;
	int			revents = 0;
	int			err;

	if (sc->faulted)
		return (ENXIO);

	err = devfs_get_cdevpriv((void **)&hc);
	if (err) {
		device_printf(sc->dev,
		    "heci_poll: devfs_get_cdevpriv failed: %d\n", err);
		return (err);
	}
	mec = hc->peer;
	if (mec == NULL) {
		device_printf(sc->dev, "heci_poll for unconnected client\n");
		return (ENOTCONN);
	}

	if (events & (POLLIN | POLLRDNORM))
		if (mec->rcount || sc->faulted || mec->peer != hc)
			revents |= events & (POLLIN | POLLRDNORM);
	if (events & (POLLOUT | POLLWRNORM))
		if (hc->flow_ctrl || sc->faulted)
			revents |= events & (POLLOUT | POLLWRNORM);
	if (revents == 0)
		selrecord(td, &mec->si);

	return (0);
}

static int
heci_ioctl(struct cdev *cdev, u_long cmd, caddr_t data, int flag,
    struct thread *td)
{
	struct heci_softc	*sc = CDEV2SOFTC(cdev);
	struct host_client	*hc;
	struct me_client	*mec;
	int			err;

	if (sc->faulted)
		return (ENXIO);

	err = devfs_get_cdevpriv((void **)&hc);
	if (err) {
		device_printf(sc->dev,
		    "heci_ioctl: devfs_get_cdevpriv failed: %d\n", err);
		return (err);
	}

	if (cmd == HECI_CONNECT) {
		struct heci_guid *guid = (struct heci_guid *)data;

		/* XXX For now allow one host client to be connected
		 * to only one ME client.
		 */
		if (hc->peer) {
			device_printf(sc->dev,
			    "attempt to connect already connected host client\n");
			return (EISCONN);
		}

		SLIST_FOREACH(mec, &sc->me_clients, next)
		if (memcmp(&mec->props.protocol_name, guid, sizeof(*guid)) == 0)
			break;

		if (mec == NULL) {
			device_printf(sc->dev,
			    "could not find ME client with given guid: ");
			print_guid(guid);
			printf("\n");
			return (ENOENT);
		}

		/* XXX For now allow only one connection from host client to one ME client.
		 * Should this be controlled by heci_client_properties.single_recv_buf ?
		 * Or, most probably, max_number_of_connections?
		 */
		if (mec->peer)
			return (EBUSY);

		HECI_LOCK(sc);
		sc->bus_req.conn.cmd.cmd = CLIENT_CONNECT_REQ_CMD;
		sc->bus_req.conn.me_addr = mec->addr;
		sc->bus_req.conn.host_addr = hc->addr;
		err = heci_bus_cmd(sc, 1);
		if (bootverbose && !err)
			device_printf( sc->dev,
			    "host 0x%02x, client 0x%02x: connect status 0x%02x\n",
			    hc->addr, mec->addr, sc->bus_resp.conn.status);
		heci_bus_release(sc);

		if (err) {
			device_printf( sc->dev,
			    "host 0x%02x, client 0x%02x: connect failed: %d\n",
			    hc->addr, mec->addr, err);
			HECI_UNLOCK(sc);
			return (err);
		}

		CLIENT_LOCK(mec);
		if (mec->peer == NULL) {
			hc->peer = mec;
			mec->peer = hc;
			CLIENT_UNLOCK(mec);
		}
		else {
			CLIENT_UNLOCK(mec);
			HECI_UNLOCK(sc);
			return (EBUSY);
		}
		HECI_UNLOCK(sc);
	}
	else {
		device_printf(sc->dev, "unknown ioctl %lu\n", cmd);
		return (EINVAL);
	}

	return (0);
}

static void
print_guid(const struct heci_guid *g)
{
	printf("%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X",
	    g->data1, g->data2, g->data3,
	    g->data4[0], g->data4[1], g->data4[2], g->data4[3],
	    g->data4[4], g->data4[5], g->data4[6], g->data4[7]);
}

static void
print_fw_client_props(struct heci_softc *sc,
	const struct heci_client_properties *props)
{
	device_printf(sc->dev, "\tprotocol_name(guid) = ");
	print_guid(&props->protocol_name);
	printf("\n");

	if (!bootverbose)
		return;

	device_printf(sc->dev, "\tprotocol_version = 0x%02x\n",
	    props->protocol_version);
	device_printf(sc->dev, "\tmax_number_of_connections = 0x%02x\n",
	    props->max_number_of_connections);
	device_printf(sc->dev, "\tfixed_address = 0x%02x\n",
	    props->fixed_address);
	device_printf(sc->dev, "\tsingle_recv_buf = 0x%02x\n",
	    props->single_recv_buf);
	device_printf(sc->dev, "\tmax_msg_length = %u\n",
	    props->max_msg_length);
}
