/*
 * Part of Intel(R) Manageability Engine Interface Linux driver
 *
 * Copyright (c) 2003 - 2008 Intel Corp.
 * All rights reserved.
 *
 * Copyright (c) 2009 Andriy Gapon
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 */

#ifndef HECI_H
#define HECI_H

/* Version of host driver API. */
#define HBM_MINOR_VERSION                   0
#define HBM_MAJOR_VERSION                   1

/*
 *  HECI BUS Interface Section
 *  All messages passed through HECI bus have this header.
 *  Note that me_addr are set by fw, host_addr must be unique for
 *  connecting host clients. host_addr of 0, 1 and 2 are reserved.
 */
struct heci_msg_hdr {
	uint32_t me_addr:8;		/* address of fw client, zero for bus messages */
	uint32_t host_addr:8;		/* address of host client, zero for bus messages */
	uint32_t length:9;		/* message length, not including this header */
	uint32_t reserved:6;
	uint32_t msg_complete:1;	/* if this is a complete or multi-part message */
} __packed;


/* All HECI *bus* messages start with a byte defined here,
 * it determines interpretation of the bus message.
 */
#define HOST_RESPONSE_MASK		    0x80
struct hbm_cmd {
	uint8_t cmd:7;
	uint8_t is_response:1;
} __packed;


struct heci_bus_message {
	struct hbm_cmd cmd;
	uint8_t command_specific_data[];
} __packed;


struct hbm_version {
	uint8_t minor_version;
	uint8_t major_version;
} __packed;


/* This bus message is sent to initiate
 * communication via HECI bus.
 */
#define HOST_START_REQ_CMD                  0x01
struct hbm_host_version_request {
	struct hbm_cmd cmd;
	uint8_t reserved;
	struct hbm_version host_version;
} __packed;


#define HOST_START_RES_CMD                  (HOST_RESPONSE_MASK|HOST_START_REQ_CMD)
struct hbm_host_version_response {
	struct hbm_cmd cmd;
	int host_version_supported;
	struct hbm_version me_max_version;
} __packed;


/* This bus message is sent to tear down and cease
 * all communication via HECI bus.
 */
#define HOST_STOP_REQ_CMD                   0x02
struct hbm_host_stop_request {
	struct hbm_cmd cmd;
	uint8_t reason;
	uint8_t reserved[2];
} __packed;


#define HOST_STOP_RES_CMD                   (HOST_RESPONSE_MASK|HOST_STOP_REQ_CMD)
struct hbm_host_stop_response {
	struct hbm_cmd cmd;
	uint8_t reserved[3];
} __packed;


/* This bus message can come from ME when it
 * wishes to cease all communication via HECI bus.
 * This can happen e.g. when ME goes to power-saving state.
 * Expected host action is to send HOST_STOP_REQ_CMD.
 */
#define ME_STOP_REQ_CMD                     0x03
struct hbm_me_stop_request {
	struct hbm_cmd cmd;
	uint8_t reason;
	uint8_t reserved[2];
} __packed;


/* This bus message is sent to discover
 * clients residing in ME firmware.
 */
#define HOST_ENUM_REQ_CMD                   0x04
struct hbm_host_enum_request {
	struct hbm_cmd cmd;
	uint8_t reserved[3];
} __packed;


#define HOST_ENUM_RES_CMD                   (HOST_RESPONSE_MASK|HOST_ENUM_REQ_CMD)
struct hbm_host_enum_response {
	struct hbm_cmd cmd;
	uint8_t reserved[3];
	uint8_t valid_addresses[32];
} __packed;


struct heci_client_properties {
	struct heci_guid protocol_name;		/* client guid */
	uint8_t protocol_version;
	uint8_t max_number_of_connections;	/* probably can't communicate if this is zero */
	uint8_t fixed_address;			/* XXX can't communicate if this is one */
	uint8_t single_recv_buf;		/* one buffer is shared for comms with all host clients */
	uint32_t max_msg_length;
} __packed;


/* This bus message is sent to query
 * properties/capabilities of a particular fw client
 * identified by 'address' field.
 */
#define HOST_CLIENT_PROPERTIES_REQ_CMD      0x05
struct hbm_props_request {
	struct hbm_cmd cmd;
	uint8_t address;
	uint8_t reserved[2];
} __packed;


#define HOST_CLIENT_PROPERTIES_RES_CMD      (HOST_RESPONSE_MASK|HOST_CLIENT_PROPERTEIS_REQ_CMD)
struct hbm_props_response {
	struct hbm_cmd cmd;
	uint8_t address;
	uint8_t status;
	uint8_t reserved[1];
	struct heci_client_properties client_properties;
} __packed;


/* This bus message is sent to establish
 * connection between a particular host client and a particular
 * fw client identified by 'host_addr' and 'me_addr'.
 * Note that in addition to hbm_client_connect_response generated
 * by heci bus implementation in fw, the fw client may also generate
 * hbm_flow_control message to signal that it is ready to receive
 * input message.
 */
#define CLIENT_CONNECT_REQ_CMD              0x06
struct hbm_client_connect_request {
	struct hbm_cmd cmd;
	uint8_t me_addr;
	uint8_t host_addr;
	uint8_t reserved;
} __packed;


#define CLIENT_CONNECT_RES_CMD              (HOST_RESPONSE_MASK|CLIENT_CONNECT_REQ_CMD)
struct hbm_client_connect_response {
	struct hbm_cmd cmd;
	uint8_t me_addr;
	uint8_t host_addr;
	uint8_t status;
} __packed;


#define CLIENT_DISCONNECT_REQ_CMD           0x07
struct hbm_client_disconnect_request {
	struct hbm_cmd cmd;
	uint8_t me_addr;
	uint8_t host_addr;
	uint8_t reserved[1];
} __packed;


#define CLIENT_DISCONNECT_RES_CMD           (HOST_RESPONSE_MASK|CLIENT_DISCONNECT_REQ_CMD)
/* note: this is precisely the same as hbm_client_connect_response */
struct hbm_client_disconnect_response {
	struct hbm_cmd cmd;
	uint8_t me_addr;
	uint8_t host_addr;
	uint8_t status;
} __packed;


/* This bus message is sent by either
 * connected host or connected firmware to indicate that
 * it is ready to receive a message from its peer.
 * E.g. host client should receive at least one such message before
 * it can send an actual request to firmware client, if the host
 * client expects to get a response it should also send a flow
 * control message.
 */
#define HECI_FLOW_CONTROL_CMD               0x08
struct hbm_flow_control {
	struct hbm_cmd cmd;
	uint8_t me_addr;
	uint8_t host_addr;
#define HECI_FC_MESSAGE_RESERVED_LENGTH 5
	uint8_t reserved[HECI_FC_MESSAGE_RESERVED_LENGTH];
} __packed;


/* Utility types. */
union hbm_any_request {
	struct hbm_cmd cmd;
        struct hbm_host_version_request vers;
        struct hbm_host_stop_request stop;
        struct hbm_host_enum_request enumer;
        struct hbm_props_request props;
        struct hbm_client_connect_request conn;
        struct hbm_client_disconnect_request disconn;
        struct hbm_flow_control flow; /* can be sent both ways */
};

union hbm_any_response {
	struct hbm_cmd cmd;
        struct hbm_host_version_response vers;
        struct hbm_host_stop_response stop;
        struct hbm_me_stop_request me_stop; /* this request comes from ME */
        struct hbm_client_disconnect_request disconn; /* ME may request disconnect */
        struct hbm_host_enum_response enumer;
        struct hbm_props_response props;
        struct hbm_client_connect_response conn; /* and the same for disconnect */
        struct hbm_flow_control flow; /* can be sent both ways */
};

union hbm_any_msg {
	union hbm_any_request req;
	union hbm_any_response resp;
};

#endif /* HECI_H */
